/* peggy -- a JPEG viewer for GGI 
 *   Version 0.3
 *
 * Copyright 1997 Andreas Vogler
 *
 * This code is under GPL. Read LICENSE for more information
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <asm/page.h>
#include <jpeglib.h>
#include <ggi/ggi_commands.h>

#undef DEBUG

int ggi_fd;
unsigned int time_to_sleep=5;
unsigned int num_pics=0;

struct resolution {
	int x,y;
} resls [] =
  {
	  320,200,
	  640,480,
	  800,600,
	  1024,768,
	  1280,1024,
	  0,0
  };

ggi_mode * ggi_graphmode(int xsize,int ysize,
			 int xvirtual,int yvirtual,int type)
{ 
	int err;
	static ggi_mode tim={ 1,
		       { 800, 600},
		       {1024,1024},
		       { 300, 220},
		       GT_8BIT,
		       {8,16},
		       /*0*/      };
	tim.visible.x=xsize;
	tim.visible.y=ysize;
	tim.virtual.x=xvirtual;
	tim.virtual.y=yvirtual;                       
	tim.graphtype=type;

	err=ioctl(ggi_fd,DRIVER_SETGRAPHMODE,&tim);
	
	if (err<0)
		return NULL;
	
	err=ioctl(ggi_fd,DRIVER_GETGRAPHMODE,&tim);
	if (err<0)
		return NULL;

#if 0	
	printf("Mode set: (%dx%d %d)\n",
	       tim.virtual.x,tim.virtual.y,tim.graphtype);
#endif	
	return &tim;
}

int ggi_testmode(int xsize,int ysize,
		 int xvirtual,int yvirtual,int type)
{ 
	int err;
	ggi_mode tim={ 1,
		       { 800, 600},
		       {1024,1024},
		       { 300, 220},
		       GT_8BIT,
		       {8,16},
		       /*0*/      };
	tim.visible.x=xsize;
	tim.visible.y=ysize;
	tim.virtual.x=xvirtual;
	tim.virtual.y=yvirtual;                       
	tim.graphtype=type;

	err=ioctl(ggi_fd,DRIVER_TESTMODE,&tim);
	
	if (err<0)
		return err;
	
	return 0;
}

void do_jpeg(const char * file, int do_dither, int width, int height)
{
	FILE * infile;
	struct jpeg_decompress_struct cinfo;
	struct jpeg_error_mgr jerr;
	ggi_color * cl=NULL;
	ggi_mode * mode;
	int i;
	unsigned int x,y,xoff,yoff,size;
	unsigned char * graphics_mem=NULL, * p;
	unsigned char * buf=NULL;
	int grmode;

#ifdef DEBUG
	printf("opening file: %s\n", file);
#endif

	infile=fopen(file, "rb");
	if(!infile) {
		printf("opening file failed!!\n");
		return;
	}

	if(do_dither)
		grmode=GT_8BIT;
	else
		grmode=GT_32BIT;

	cinfo.err=jpeg_std_error(&jerr);
	jpeg_create_decompress(&cinfo);
	jpeg_stdio_src(&cinfo, infile);
	jpeg_read_header(&cinfo, TRUE);

	if(do_dither) {
		cinfo.quantize_colors=TRUE;
		cinfo.dither_mode=JDITHER_FS;
		cinfo.desired_number_of_colors=255;
	}

	jpeg_calc_output_dimensions(&cinfo);

	/* find out the best resolution and scaling values */

#ifdef DEBUG
	printf("testing for best resolution and scaling\n");
	printf("starting with %ix%i\n", cinfo.output_width,
		cinfo.output_height);
#endif

	if(width) {		/* use specified geometry */
		if(ggi_testmode(width,height,width,height,grmode)<0) {
			printf("could not set graphics mode!\n");
			jpeg_destroy_decompress(&cinfo);
			fclose(infile);
			return;
		}
		while(cinfo.output_width>width
		      || cinfo.output_height>height) {
			cinfo.scale_denom++;
			jpeg_calc_output_dimensions(&cinfo);
		}
		x=width;
		y=height;
	} else {
		i=0;
		while(resls[i].x!=0) {
			if(cinfo.output_width<=resls[i].x
			   && cinfo.output_height<=resls[i].y)
				break;
			i++;
		}
		if(resls[i].x==0)
			i--;
		while(i>=0) {
			if(ggi_testmode(resls[i].x, resls[i].y,
				     resls[i].x, resls[i].y,
				     grmode) == 0)
				break;
			i--;
		}
		if(i<0) {
			printf("could not set graphics mode!\n");
			jpeg_destroy_decompress(&cinfo);
			fclose(infile);
			return;
		}
		x=resls[i].x;
		y=resls[i].y;
		while(cinfo.output_width>x
		      || cinfo.output_height>y) {
			cinfo.scale_denom++;
			jpeg_calc_output_dimensions(&cinfo);
		}
	}

	/* set graphics mode */
	mode=ggi_graphmode(x,y,x,y,grmode);
	x=mode->virtual.x;
	y=mode->virtual.y;

	if(do_dither)
		size = x * y;
	else
		size = x * y * 4;

	/* calculates mmap size as in libggi */
	size = ((size >> 16) + ((size & 0xffff) ? 1: 0)) << 16;

	graphics_mem=mmap((caddr_t)0,		/* Map whereever you like   */
			  size,
			  PROT_WRITE,		/* for reading and writing  */
			  MAP_SHARED,		/* shared mapping */
			  ggi_fd,		/* from the graphics-device */
			  MMAP_TYPE_MMIO |
			  MMAP_PER_REGION_TYPE | MMAP_FRAMEBUFFER );

	if(graphics_mem == (void *)-1) {
		printf("mmap failed!!\n");
		jpeg_destroy_decompress(&cinfo);
		fclose(infile);
		return;
	}

	p=graphics_mem;

	/* clear framebuffer */

	if(do_dither) {
		for(i=0; i<x*y; i++) {
			graphics_mem[i]=0;
		}
	} else {
		for(i=0; i<x*y*4; i++) {
			graphics_mem[i]=0;
		}
	}

	/* buffer to store one row of image data */
	buf=(unsigned char *)malloc(cinfo.output_width 
				    * cinfo.output_components);
	if(!buf) {
		printf("memory allocation failed!!\n");
		jpeg_destroy_decompress(&cinfo);
		fclose(infile);
		exit(1);
	}

	jpeg_start_decompress(&cinfo);

	/* center picture */
	xoff = (x - cinfo.output_width);	
	yoff = (y - cinfo.output_height);

	if(do_dither) {
		/* set palette */

		cl=(ggi_color *)malloc(cinfo.actual_number_of_colors
				       * sizeof(ggi_color));
		if(!cl) {
			printf("memory allocation failed!!\n");
			jpeg_destroy_decompress(&cinfo);
			fclose(infile);
			free(buf);
			exit(1);
		}
		for(i=0; i<cinfo.actual_number_of_colors; i++) {
			cl[i].r=cinfo.colormap[0][i] << 8;
			cl[i].g=cinfo.colormap[1][i] << 8;
			cl[i].b=cinfo.colormap[2][i] << 8;
		}
		if(ioctl(ggi_fd, RAMDAC_SETCLUT, cl)<0) {
			printf("cannot set colormap\n");
		}

		p += ((yoff/2 * x) + xoff/2); 
		
		while(cinfo.output_scanline < cinfo.output_height) {
			jpeg_read_scanlines(&cinfo, &buf, 1);
			for(i=0; i<cinfo.output_width; i++) {
				*p =buf[i];
				p++;
			}
			p += xoff;
		}
	} else {
		p += ((yoff/2 * x) + xoff/2) * 4; 
		
		while(cinfo.output_scanline < cinfo.output_height) {
			jpeg_read_scanlines(&cinfo, &buf, 1);
			for(i=0; i<cinfo.output_width; i++) {
				*(unsigned int *)p = 
					buf[i*3] << 16 | buf[i*3+1] << 8 | buf[i*3+2];
				p += 4;
			}
			p += xoff * 4;
		}
	}

	jpeg_finish_decompress(&cinfo);
		
	jpeg_destroy_decompress(&cinfo);

	sleep(time_to_sleep);

	/* clear framebuffer */
	if(do_dither) {
		for(i=0; i<x*y; i++) {
			graphics_mem[i]=0;
		}
	} else {
		for(i=0; i<x*y*4; i++) {
			graphics_mem[i]=0;
		}
	}

	munmap(graphics_mem, size);

	free(cl);
	free(buf);
	fclose(infile);
}

void usage(const char * n)
{
	printf("
USAGE:

%s [--geometry XXXxYYY] [--dither] [--time sec] [--random] 
   [--version] [--help] <filename> [<filename> ... ]

abbrevetions allowed (e.g. -d, -t, ... or --rand, ...)

You must use dither to view on an 8 bit display.

", n);
}

void version_info()
{
	printf("
Peggy -- simple JPEG viewer for GGI
Version: 0.3

Copyright 1997 Andreas Vogler

Peggy is free software and you are welcome to distribute copies of
it under certain conditions. Read the GNU General Public License for
further information.

");
}

unsigned int choosep()
{
	static unsigned int * mem=NULL;
	unsigned int num;
	
	if(!mem) {
		mem=(unsigned int *)malloc(num_pics/32 + (num_pics%32) ? 1:0);
		if(!mem) {
			printf("out of memory!\n");
			exit(1);
		}
	}

	do {
		num=(unsigned int)((float)rand()/(float)RAND_MAX*num_pics);
	}
	while (mem[num >> 5] & (1 << (num & 0x1f)));
	mem[num >> 5] |= (1 << (num & 0x1f));
	return num;
}

void main(int argc, char * argv[])
{
	int i=1, count, dither=0;
	int random=0, x=0, y=0;

	while(i<argc && *argv[i]=='-') {
		if(strncmp("-d", argv[i], 2)==0 ||
		   strncmp("--d", argv[i], 3)==0) {
			dither=1;
			i++;
			continue;
		}
		if(strncmp("-t", argv[i], 2)==0 ||
		   strncmp("--t", argv[i], 3)==0) {
			time_to_sleep=strtoul(argv[i+1],NULL,0);
			i+=2;
			continue;
		}
		if(strncmp("-r", argv[i], 2)==0 ||
		   strncmp("--r", argv[i], 3)==0) {
			random=1;
			i++;
			continue;
		}
		if(strncmp("-g", argv[i], 2)==0 ||
		   strncmp("--g", argv[i], 3)==0) {
			sscanf(argv[i+1], "%ix%i", &x, &y);
			if((x+y) && (!x || !y)) {
				printf("invalid values for geometry\n");
				x=0; y=0;
			}
			i+=2;
			continue;
		}
		if(strncmp("-v", argv[i], 2)==0 ||
		   strncmp("--v", argv[i], 3)==0) {
			version_info();
			i++;
			continue;
		}
		if(strncmp("-?", argv[i], 2)==0 ||
		   strncmp("-h", argv[i], 2)==0 ||
		   strncmp("--h", argv[i], 3)==0) {
			version_info();
			usage(argv[0]);
			i++;
			continue;
		}
		printf("unrecognized option: %s\n", argv[i]);
		usage(argv[0]);
		i++;
	}
	
	num_pics=argc-i;
	if(num_pics==0)
		exit(0);

#ifdef DEBUG
	printf("opening /dev/graphic\n");
#endif

	ggi_fd=open("/dev/graphic",O_RDWR);
	if(ggi_fd==-1) {
		printf("could not access KGI module\n");
		exit(1);
	}

	if(random) {
		srand(time(0));
		for(count=0; count <  num_pics; count++) {
			do_jpeg(argv[i+choosep()], dither, x, y);
		}
	} else {
		for(;i<argc;i++) {
			do_jpeg(argv[i],dither, x, y);
		}
	}

	close(ggi_fd);
}
